#define RC__USE_VLC
#define RC__USE_MPC
#define RC__USE_WMP
#include "usb_interface.h"
#include <fstream>

#define RESPONSE_TIME 600

static const wchar_t VERSION_NUMBER[] = L"1.22";

using namespace std;

LPDIRECTINPUT8 lpdi        = NULL;		//DirectInput Manager Object
LPDIRECTINPUTDEVICE8 lpRec = NULL;		//Device Object for HID remote

static bool done = false;
static const wchar_t wclassname[] = L"Remote Controller";
static char* winamp_path = NULL;		//Stores the path to winamp.exe

void ButtonPressed(int id)
{
	static int timestamp = 0;
	if(id == 99) { timestamp = 0; return; }
	if(GetTickCount() - timestamp <= RESPONSE_TIME)
		return;

#ifdef RC__USE_MPC
	if(SendMediaPlayerClassicMessage(MPC_NIL)) { return; }
#endif
#ifdef RC__USE_VLC
	if(SendVLCMessage(id)) { timestamp = GetTickCount();  return; }
#endif

	switch(id) {
		case 4:   //FFW
			SendWinampMessage(winampAPI.WA_FFW);
			break;
		case 5:   //FBW
			SendWinampMessage(winampAPI.WA_FRW);
			break;
		case 6:   //Info/EPG
			WinExec("scrnsave.scr /s", SW_SHOWDEFAULT);
			break;
		case 11:  //Stop
			SendWinampMessage(winampAPI.WA_STOP);
			break;
		case 13:  //Back
			break;
		case 14:  //PgDown
			SendWinampMessage(winampAPI.WA_VOLUMEDOWN);
			return;		//return here to avoid reset of timestamp
		case 15:  //PgUp
			SendWinampMessage(winampAPI.WA_VOLUMEUP);
			return;		//return here to avoid reset of timestamp
		case 16:  //Rec
			break;
		case 17:  //+
			break;
		case 18:  //-
			break;
		case 19:  //Mute
			break;
		case 20:  //Play
			if(!SendWinampMessage(winampAPI.WA_MYPLAY))
				//WinExec("c:\\programme\\winamp\\winamp.exe", SW_SHOWDEFAULT);
				if(winamp_path) WinExec(winamp_path, SW_SHOWDEFAULT);
			break;
		case 21:  //FM Radio
			SendWinampMessage(winampAPI.WA_PLSTART);
			break;
		case 22:  //SkipBW
			SendWinampMessage(winampAPI.WA_PREV);
			break;
		case 23:  //SkipFW
			SendWinampMessage(winampAPI.WA_NEXT);
			break;
		default:
			wchar_t buffer[255];
			swprintf_s(buffer, 255, L"New Button found!\n%d", id);
			MessageBox(NULL, buffer, wclassname, MB_OK|MB_ICONEXCLAMATION);
			break;
	}
	timestamp = GetTickCount();
}

BOOL CALLBACK DIDevEnumObjectsCallback(LPCDIDEVICEOBJECTINSTANCE lpddoi, LPVOID pvRef)
{
	if(lpddoi->guidType == GUID_XAxis) {
		MessageBox(NULL, L"X Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_YAxis) {
		MessageBox(NULL, L"Y Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_ZAxis) {
		MessageBox(NULL, L"Z Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_RxAxis) {
		MessageBox(NULL, L"RotX Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_RyAxis) {
		MessageBox(NULL, L"RotY Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_RzAxis) {
		MessageBox(NULL, L"RotZ Axis", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_Slider) {
		MessageBox(NULL, L"Slider", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_Button) {
		MessageBox(NULL, L"Button", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_Key) {
		MessageBox(NULL, L"Key", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_POV) {
		MessageBox(NULL, L"POV", L"Enum", MB_OK);
	} else if(lpddoi->guidType == GUID_Unknown) {
		MessageBox(NULL, L"Unknown", L"Enum", MB_OK);
	} else {
		MessageBox(NULL, L"Error", L"Enum", MB_OK);
	}
	return DIENUM_CONTINUE;
}

void InitLog() {
	ofstream logfile("error.log", ios_base::out|ios_base::trunc);
	if(!logfile) return;
	logfile << "Remote Controller - New Session\n";
	logfile << "--------------------------------------\n";
	logfile.close();
}

void cLog(const char* str) {
	ofstream logfile("error.log", ios_base::out|ios_base::app);
	if(!logfile) return;
	logfile << str << endl;
	logfile.close();
}

BOOL CALLBACK DIEnumCallback(const DIDEVICEINSTANCE* pInst, VOID* pContext )
{
	//Find the HID device based on the product GUID
	static bool takefirst = true;
	wchar_t buffer[512];

	BYTE devType    = (BYTE)(pInst->dwDevType & 0xFF);
	BYTE devSubType = (BYTE)((pInst->dwDevType >> 8) & 0xFF);

	//DI8DEVTYPE_DEVICE
	if(devType == DI8DEVTYPE_DEVICE) {
		swprintf_s(buffer, 512, L"Found Device:\n%s\nGUID: %08lX-%04hX-%04hX-%02X%02X-%02X%02X%02X%02X%02X%02X", pInst->tszProductName, 
			pInst->guidProduct.Data1, pInst->guidProduct.Data2, pInst->guidProduct.Data3,
			pInst->guidProduct.Data4[0], pInst->guidProduct.Data4[1], pInst->guidProduct.Data4[2], pInst->guidProduct.Data4[3],
			pInst->guidProduct.Data4[4], pInst->guidProduct.Data4[5], pInst->guidProduct.Data4[6], pInst->guidProduct.Data4[7]);
		//MessageBox(NULL, buffer, L"Enumeration", MB_OK);
		if(pInst->guidProduct.Data1 == 3238069357) {
			if(!takefirst) { takefirst = true; return DIENUM_CONTINUE; }
			//MessageBox(NULL, buffer, L"Enumeration", MB_OK);
			if( lpdi->CreateDevice(pInst->guidInstance, &lpRec, NULL) != DI_OK )
				return DIENUM_CONTINUE;
			return DIENUM_STOP;
		}
	}
	return DIENUM_CONTINUE;
}

void GetWinampPath()		//acquire winamp's path from registry
{
	HKEY hKCU;
	REGSAM sam = KEY_READ;
	LONG res;
	char buffer[1024];
	res = RegOpenCurrentUser(sam, &hKCU);
	if(res != ERROR_SUCCESS) {
		FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, NULL, res, 0, buffer, 1024, NULL);
		cLog("WARNING! - Accessing Registry failed"); cLog(buffer); return;
	}
	long len = 1024;
	res = RegQueryValueA(hKCU, "Software\\Winamp", buffer, &len);
	if(res != ERROR_SUCCESS) {
		FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, NULL, res, 0, buffer, 1024, NULL);
		cLog("WARNING! - Winamp was not found"); cLog(buffer); return;
	}
	winamp_path = new char[len+11];		//freed in WinMain()
	if(!winamp_path) { cLog("WARNING! - Out of memory"); return; }
	sprintf_s(winamp_path, len+11, "%s\\winamp.exe", buffer);
}

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch (msg) {
		case WM_KEYDOWN: {
			switch(wParam) {
				case VK_ESCAPE:
					PostMessage(hWnd, WM_CLOSE, 0, 0); return 0;
			}
		} break;
		case WM_DESTROY: {
			done = true;
			PostQuitMessage (0) ;
			return 0 ;
		} break;
		case WM_USER+2707:	//tray icon
			if(LOWORD(lParam) == WM_RBUTTONUP) {
				if(MessageBox(NULL, L"Quit?", wclassname, MB_YESNO|MB_ICONQUESTION) == 6) {
					PostMessage(hWnd, WM_CLOSE, 0, 0); return 0;
				}
			}
			break;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT) {
	InitLog();
	//ReadConfig();
	HRESULT hres;
	//Init Win32:
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), (CS_HREDRAW | CS_VREDRAW) , MsgProc, 0, 0, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      wclassname, NULL };
    RegisterClassEx(&wc);
	HWND hWnd = CreateWindow(wclassname, TEXT("Remote Controller Window"), 
                              WS_MINIMIZE, 0, 0, 50, 50,
                              NULL, NULL, wc.hInstance, NULL);

	//UpdateWindow(hWnd); ShowWindow(hWnd, SW_SHOWDEFAULT);

	//Init DI:
	if( DirectInput8Create(wc.hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8, (LPVOID*)&lpdi, NULL) != DI_OK ) {
		cLog("ERROR! - Unable to init DirectInput");
		return 1;
	}
	
	if( lpdi->EnumDevices(DI8DEVCLASS_ALL, DIEnumCallback, 0, DIEDFL_ATTACHEDONLY) != DI_OK ) {
		cLog("ERROR! - HID device not found");
		lpdi->Release();
		return 1;
	}
	if(!lpRec) { cLog("ERROR! - Unable to init HID device"); lpdi->Release(); return 1; }
	//if(lpRec->EnumObjects(DIDevEnumObjectsCallback, 0, DIDFT_ALL) != DI_OK) { lpRec->Release(); lpdi->Release(); return 1; }
	hres = lpRec->SetDataFormat(&g_didf);
	if(hres != DI_OK) { cLog("ERROR! - Unable to set Data Format"); lpRec->Release(); lpdi->Release(); return 1; }
	if(lpRec->SetCooperativeLevel(hWnd, DISCL_NONEXCLUSIVE|DISCL_BACKGROUND) != DI_OK) { cLog("ERROR! - Unable to set cooperative level (device busy?)"); lpRec->Release(); lpdi->Release(); return 1; }

	GetWinampPath();

	//Setup Task Icon:
	NOTIFYICONDATA icondata;
	icondata.cbSize = sizeof(NOTIFYICONDATA);
	icondata.hWnd = hWnd;
	icondata.uID = 1;
	icondata.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	icondata.uCallbackMessage = WM_USER + 2707;
	swprintf(icondata.szTip, 64, L"Remote Controller V-%s", VERSION_NUMBER);
	icondata.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON1));
	Shell_NotifyIcon(NIM_ADD, &icondata);

	//MAIN LOOP
	MSG msg;
	done = false;
	while(!done)
	{
		PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);
		TranslateMessage(&msg);
		DispatchMessage(&msg);

		lpRec->Poll();

		g_remote input;
		for(int i=0; i<24; i++) { input.bButton[i] = 0; }
		hres = lpRec->GetDeviceState(sizeof(g_remote), &input);
		if(FAILED(hres)) {
			hres = lpRec->Acquire();
			while(hres == DIERR_INPUTLOST) lpRec->Acquire();

			if( hres == DIERR_OTHERAPPHASPRIO || hres == DIERR_NOTACQUIRED ) {
				cLog("ERROR! Device lost");
				MessageBox(hWnd, L"Device Lost.\nExiting.", wclassname, MB_ICONERROR|MB_OK);
				done = true;  break;
			}
		}

		for(int i=0; i<24; i++) {
			if(input.bButton[i] != 0) {
				ButtonPressed(i);
				break;
			}
			if(i >= 23) ButtonPressed(99);
		}
		Sleep(80);
    }
	lpRec->Unacquire();
	
	if(lpRec) {	lpRec->Release(); lpRec = NULL; }
	if(lpdi)  { lpdi->Release();  lpdi  = NULL; }
	if(winamp_path) { delete[] winamp_path; winamp_path = NULL; }
	Shell_NotifyIcon(NIM_DELETE, &icondata);
	UnregisterClass(wclassname, wc.hInstance);
	cLog("Exiting.");
	return 0;
}